import logging

from typing import Dict, Union

from io import BytesIO
from contextlib import contextmanager

from ..layouts import LTPage
from ..layouts import LTItem
from ..layouts import LTCurve
from ..layouts import LTLine
from ..layouts import LTRect
from ..layouts import LTFigure
from ..layouts import LTImage
from ..layouts import LTChar, LTCharBlock
from ..utils import enc

from .PDFConverter import PDFConverter
from ..interpreter.PDFGraphicState import PDFGraphicStateColor

log = logging.getLogger(__name__)

def get_color(col: PDFGraphicStateColor):
    color_type, color_val = col.get_color()
    if color_type is not None:
        if color_type == 'rgb':
            return f"rgb({','.join(map(str, color_val))})"
        elif color_type == 'cmyk':
            return f"cmyk({','.join(map(str, color_val))})"
        elif color_type == 'gray':
            return f'gray({color_val})'
    return ""

class JSONConverter(PDFConverter):
    def __init__(self, 
        rsrcmgr, 
        outfp: BytesIO, 
        codec: str='utf-8', 
        pageno: int=1, 
        imagewriter=None,
    ):
        PDFConverter.__init__(self, rsrcmgr, outfp, imagewriter=imagewriter, codec=codec, pageno=pageno)
        self.__last_levels_deep:int = None
        self.__levels_deep = 0
        self.write_header()
    
    @property
    def _levels_deep(self): return self.__levels_deep

    @_levels_deep.setter
    def _levels_deep(self, val: int):
        self.__last_levels_deep = self.__levels_deep
        self.__levels_deep = val

    def write(self, text: str):
        if self.__last_levels_deep == self._levels_deep:
            text = ',' + text
        self.write_raw(text)
    
    def write_raw(self, text):
        if self.codec:
            text = text.encode(self.codec)
        self.outfp.write(text)
        self._levels_deep = self._levels_deep
    
    def write_header(self):
        self.write("{")
        self._levels_deep += 1
        if self.codec:
            self.write('"encoding": "%s"' % self.codec)
        self.write('"pages": [')
        self._levels_deep += 1
    
    def write_footer(self):
        self._levels_deep -= 1
        self.write(']')
        self._levels_deep -= 1
        self.write('}')

    @contextmanager
    def place_elm_with_child(self, tag_name: str, attr_dict: Dict[str, str] = {}):
        attrs = ", ".join([
            f'"{name}": "{value}"'
            for name, value in attr_dict.items()
        ])
        self.write('{ "type": "%s", "attrs": { %s } , "children": [' % (tag_name, attrs) )
        self._levels_deep += 1
        yield
        self._levels_deep -= 1
        self.write("]}")

    def place_elm_close(self, tag_name: str, attr_dict: Dict[str, str] = {}):
        attrs = ", ".join([
            f'"{name}": "{value}"'
            for name, value in attr_dict.items()
        ])
        self.write('{ "type": "%s", "attrs": { %s } }' % (tag_name, attrs) ) 

    def place_image(self, item: LTImage):
        name = item.name

        if self.imagewriter is not None:
            name = self.imagewriter.export_image(item)

        self.place_elm_close('img', {
            'src': enc(name, None),
            "x0": item.x0,
            "x1": item.x1,
            "y0": item.y0,
            "y1": item.y1,
        })

    def render_curve(self, item: LTCurve):
        attr = {
            "x0": item.x0,
            "x1": item.x1,
            "y0": item.y0,
            "y1": item.y1,
            "stroke-color": get_color(item.stroke),
            "fill-color": get_color(item.fill),
        }
        if isinstance(item, LTRect):
            return self.place_elm_close('rect', attr)
        elif isinstance(item, LTLine):
            return self.place_elm_close('line', attr)

        with self.place_elm_with_child('curve', attr):
            for path in item.paths:
                with self.place_elm_with_child('path', { "type": path.method.name }):
                    for point in path.points:
                        self.place_elm_close('point', { 'x': point.x, 'y': point.y })

    def receive_layout(self, ltpage: LTPage):
        def render(item: LTItem):
            if isinstance(item, LTPage):
                with self.place_elm_with_child('page', { 
                    "width": item.width,
                    "height": item.height
                }):
                    for child in item:
                        render(child)
            elif isinstance(item, LTCurve):
                self.render_curve(item)
            elif isinstance(item, LTFigure):
                with self.place_elm_with_child('figure', { 
                    "x0": item.x0,
                    "x1": item.x1,
                    "y0": item.y0,
                    "y1": item.y1,
                }):
                    for child in item:
                        render(child)
            elif isinstance(item, LTImage):
                self.place_image(item)
            elif isinstance(item, LTChar):
                self.place_elm_close('char', {
                    'font-size': item.size,
                    'font-family': item.fontname,
                    "x0": item.x0,
                    "x1": item.x1,
                    "y0": item.y0,
                    "y1": item.y1,
                    "stroke-color": get_color(item.graphicstate.scolor),
                    "fill-color": get_color(item.graphicstate.ncolor),
                    "text": item.get_text()
                })
            elif isinstance(item, LTCharBlock):
                self.place_elm_close('char-block', {
                    'font-size': item.size,
                    'font-family': item.fontname,
                    "x0": item.x0,
                    "x1": item.x1,
                    "y0": item.y0,
                    "y1": item.y1,
                    "stroke-color": get_color(item.graphicstate.scolor),
                    "fill-color": get_color(item.graphicstate.ncolor),
                    "text": item.get_text()
                })
        render(ltpage)

    def close(self):
        self.write_footer()
    


import logging

from .cmapdb import CMapDB, CMap

from .PDFFont import PDFFontError
from .PDFFont import PDFType1Font
from .PDFFont import PDFTrueTypeFont
from .PDFFont import PDFType3Font
from .PDFFont import PDFCIDFont

from .PDFStream import PDFException
from .PDFStream import resolve1, dict_value, list_value
from .constants import LITERAL_PDF, LITERAL_TEXT, LITERAL_FONT
from .PSStackParser import literal_name, settings

log = logging.getLogger(__name__)

##  Exceptions
##
class PDFResourceError(PDFException):
    pass

class PDFInterpreterError(PDFException):
    pass

##  Resource Manager
##
class PDFResourceManager(object):

    """Repository of shared resources.

    ResourceManager facilitates reuse of shared resources
    such as fonts and images so that large objects are not
    allocated multiple times.
    """

    def __init__(self, caching=True):
        self.caching = caching
        self._cached_fonts = {}
        return

    def get_procset(self, procs):
        for proc in procs:
            if proc is LITERAL_PDF:
                pass
            elif proc is LITERAL_TEXT:
                pass
            else:
                #raise PDFResourceError('ProcSet %r is not supported.' % proc)
                pass
        return

    def get_cmap(self, cmapname, strict=False):
        try:
            return CMapDB.get_cmap(cmapname)
        except CMapDB.CMapNotFound:
            if strict:
                raise
            return CMap()

    def get_font(self, objid: int, spec: dict):
        if objid and objid in self._cached_fonts:
            font = self._cached_fonts[objid]
        else:
            log.info('get_font: create: objid=%r, spec=%r', objid, spec)
            if settings.STRICT:
                if spec['Type'] is not LITERAL_FONT:
                    raise PDFFontError('Type is not /Font')
            # Create a Font object.
            if 'Subtype' in spec:
                subtype = literal_name(spec['Subtype'])
            else:
                if settings.STRICT:
                    raise PDFFontError('Font Subtype is not specified.')
                subtype = 'Type1'
            if subtype in ('Type1', 'MMType1'):
                # Type1 Font
                font = PDFType1Font(self, spec)
            elif subtype == 'TrueType':
                # TrueType Font
                font = PDFTrueTypeFont(self, spec)
            elif subtype == 'Type3':
                # Type3 Font
                font = PDFType3Font(self, spec)
            elif subtype in ('CIDFontType0', 'CIDFontType2'):
                # CID Font
                font = PDFCIDFont(self, spec)
            elif subtype == 'Type0':
                # Type0 Font
                dfonts = list_value(spec['DescendantFonts'])
                assert dfonts
                subspec = dict_value(dfonts[0]).copy()
                for k in ('Encoding', 'ToUnicode'):
                    if k in spec:
                        subspec[k] = resolve1(spec[k])
                font = self.get_font(None, subspec)
            else:
                if settings.STRICT:
                    raise PDFFontError('Invalid Font spec: %r' % spec)
                font = PDFType1Font(self, spec)  # this is so wrong!
            if objid and self.caching:
                self._cached_fonts[objid] = font
        return font



from typing import List, Tuple

INF = (1<<31) - 1
Point = Tuple[float]
unicode = str

class Bbox:
    """A generic boundary box
    """

    def __init__(self, x0: float, y0: float, x1: float, y1: float):
        self.x0 = x0
        self.x1 = x1
        self.y0 = y0
        self.y1 = y1
    
    @property
    def height(self) -> float:
        return self.y1 - self.y0
    
    @property
    def width(self) -> float:
        return self.x1 - self.x0

    def __str__(self):
        return '%.3f,%.3f,%.3f,%.3f' % (self.x0, self.y0, self.x1, self.y1)
    
    def __repr__(self):
        return f"<Bbox {str(self)}/>"

    @classmethod
    def from_points(cls, pts: List[Point]):
        """Converts a list of points to a Boundary box."""
        (x0, y0, x1, y1) = (INF, INF, -INF, -INF)
        for (x, y) in pts:
            x0 = min(x0, x)
            y0 = min(y0, y)
            x1 = max(x1, x)
            y1 = max(y1, y)
        return Bbox(x0, y0, x1, y1)


# ==============
# Dependencies
# ==============
from setuptools import setup, find_packages
import sys

# ==============
# PARAMS
# ==============
with open('README.md') as readme_file:
    SETUP = {
        "name": "pdfmajor",
        "version": "1.1.5",
        "packages": find_packages(include=['pdfmajor*']),
        "install_requires": [
            'pycryptodome', 
            'chardet'
        ],
        "description": 'PDF parser',
        "long_description": readme_file.read(),
        "license": 'MIT/X',
        "author": 'Ariel Sosnovsky + Yusuke Shinyama + Philippe Guglielmetti',
        "author_email": 'ariel@sosnovsky.ca',
        "url": 'https://github.com/asosnovsky/pdfmajor',
        "scripts": [
            'tools/pdfconvert.py',
        ],
        "keywords": [
            'pdf parser',
            'pdf converter',
            'text mining',
        ],
        "classifiers": [
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: MIT License',
            'Topic :: Text Processing',
        ],
    }

# ==============
# Update __init__.py
# ==============
from bin.recreate_init import update_pdfmajor__init___, update_current_version_lock
update_pdfmajor__init___(SETUP["version"], SETUP["long_description"])
update_current_version_lock(SETUP["version"])

# ==============
# Run Setup
# ==============
setup(**SETUP)



import logging
import re
import os

from typing import List
from io import BytesIO, StringIO, TextIOWrapper

from ..layouts import PDFLayoutAnalyzer
from ..interpreter.PDFResourceManager import PDFResourceManager
from ..interpreter.PDFPage import PDFPage
from ..interpreter.PDFPageInterpreter import PDFPageInterpreter
from ..imagewriter import ImageWriter

log = logging.getLogger(__name__)

##  PDFConverter
##
class PDFConverter(PDFLayoutAnalyzer):

    @classmethod
    def parse_file(cls, 
        input_file: TextIOWrapper, 
        output_file: TextIOWrapper, 
        image_folder_path: str = None,
        codec: str = 'utf-8',
        maxpages: int = 0, 
        password: str = None, 
        caching: bool = True, 
        check_extractable: bool = True,
        pagenos: List[int] = None,
        dont_export_images: bool = False,
        debug: bool = False, 
    ) -> TextIOWrapper:
        
        if debug:
            log.setLevel(logging.DEBUG)
        
        if dont_export_images:
            image_folder_path = None

        if not dont_export_images and image_folder_path is None and output_file.name is not None:
            image_folder_path = os.path.dirname(output_file.name)

        log.debug("Creating resources....")
        rsrcmgr = PDFResourceManager(caching=caching)
        imagewriter=ImageWriter(image_folder_path)
        device = cls(
            rsrcmgr, 
            output_file,
            codec=codec,
            imagewriter=imagewriter
        )
        interpreter = PDFPageInterpreter(rsrcmgr, device)
        log.debug("Reading pages....")
        pages = PDFPage.get_pages(input_file, pagenos=pagenos, maxpages=maxpages, password=password, caching=caching, check_extractable=check_extractable)
        for idx, page in enumerate(pages):
            log.debug(f"Reading page #{idx}")
            interpreter.process_page(page)    
        device.close()
        return output_file
        
    def __init__(self, rsrcmgr: PDFResourceManager, outfp: TextIOWrapper, imagewriter: ImageWriter, codec: str='utf-8', pageno: int=1):
        PDFLayoutAnalyzer.__init__(self, rsrcmgr, pageno=pageno)
        self.imagewriter = imagewriter
        self.outfp = outfp
        self.codec = codec
        if hasattr(self.outfp, 'mode'):
            if 'b' in self.outfp.mode:
                self.outfp_binary = True
            else:
                self.outfp_binary = False
        else:
            if isinstance(self.outfp, BytesIO):
                self.outfp_binary = True
            elif isinstance(self.outfp, StringIO):
                self.outfp_binary = False
            else:
                try:
                    self.outfp.write(u"é")
                    self.outfp_binary = False
                except TypeError:
                    self.outfp_binary = True
        return
